﻿using System;
using System.Reflection;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using Sage.Reporting.Engine.Integration;

namespace Sage200ThreadLogon
{
    /// <summary>
    /// Provides a library class for printing a reporting using the Sage reporting engine.
    /// </summary>
    public class ReportExporter : IReportingEngineHost
    {
        //representation of the Report Designer
        private IReportingEngineSite _site;

        /// <summary>
        /// Print/Exports a Sage200 report using standard report Criteria.
        /// </summary>
        /// <param name="reportFileName">The short path and file to the report. This shouldn't be the full path, just the stub and file name e.g. Layouts\Report.layout</param>
        /// <param name="criteria">The name value pairs for the report criteria.</param>
        /// <param name="exportFilePath">The file to export the report to, when empty or null the report is not exported.</param>
        /// <param name="print">Indicate to print the report to default printer.</param>
        /// <param name="printerName">The printer name to print; if null or empty the report will print to the default printer.</param>
        public ReportExporter(string reportFileName, Dictionary<string, object> criteria, string exportFilePath, bool print, string printerName)
        {
            if (!print && string.IsNullOrEmpty(exportFilePath))
            {
                return;
            }

            ReportingEngine engine = null;
            IExportService exportService = null;

            try
            {
                //instantiate Reporting Engine
                engine = new ReportingEngine();

                _site = engine.CreateInstance(this);

                exportService = (IExportService)this.Site.GetService(ServiceIdentifiers.Export);

                //set the criteria
                foreach (string criteriaKey in criteria.Keys)
                {
                    exportService.Criteria.Add(criteriaKey, criteria[criteriaKey]);
                }

                //load and run the report:
                exportService.Load(System.IO.Path.Combine(Sage.Accounting.Application.RootPath, reportFileName));

#if !SAGE2010
                ExportFlags flags = ExportFlags.SuppressUserInteraction | ExportFlags.UseDefaultPrinter;
#else
                ExportFlags flags = ExportFlags.SuppressUserInteraction;
#endif


                //Export the report when the export path is not null.
                if (!string.IsNullOrEmpty(exportFilePath))
                {
                    ExportType exportType = ExportTypeFromFileName(exportFilePath);
                    PrintReport(exportFilePath, exportService, flags, exportType, printerName);
                }

                //Print when not false.
                if (print)
                {
                    PrintReport(string.Empty, exportService, flags, ExportType.prn, printerName);
                }
            }
            finally
            {
            }
        }

        /// <summary>
        /// Print/Exports a Sage200 report using transient/datapiped data.
        /// </summary>
        /// <param name="reportFileName">The short path and file to the report. This shouldn't be the full path, just the stub and file name e.g. Layouts\Report.layout</param>
        /// <param name="reportCatalogue">The report catalogue.</param>
        /// <param name="transientData">The list/collection of data to be piped to the report.</param>
        /// <param name="exportFilePath">The file to export the report to, when empty or null the report is not exported.</param>
        /// <param name="print">Indicate to print the report to default printer.</param>
        /// <param name="printerName">The printer name to print; if null or empty the report will print to the default printer.</param>
        public ReportExporter(string reportFileName, string reportCatalogue, Sage.ObjectStore.MetaDataObjectCollection transientData, string exportFilePath, bool print, string printerName)
        {
            if (!print && string.IsNullOrEmpty(exportFilePath))
            {
                return;
            }

            //instantiate Reporting Engine
            ReportingEngine engine = new ReportingEngine();

            _site = engine.CreateInstance(this);

            IExportService exportService = (IExportService)this.Site.GetService(ServiceIdentifiers.Export);

            IList transientReportData = BuildTransientData(transientData, reportCatalogue);

            //set the criteria
            exportService.SetAttribute(reportCatalogue, transientReportData);

            //load and run the report:
            exportService.Load(System.IO.Path.Combine(Sage.Accounting.Application.RootPath, reportFileName));

            ExportFlags flags = ExportFlags.SuppressUserInteraction;

            //Export the report when the export path is not null.
            if (!string.IsNullOrEmpty(exportFilePath))
            {
                ExportType exportType = ExportTypeFromFileName(exportFilePath);
                PrintReport(exportFilePath, exportService, flags, exportType, string.Empty);
            }

            //Print when not false.
            if (print)
            {
                PrintReport(string.Empty, exportService, flags, ExportType.prn, printerName);
            }
        }

        #region IReportingEngineHost implementation

        public object GetService(Guid identifier)
        {
            return null;
        }

        public Sage.Reporting.Engine.Integration.IReportingEngineSite Site
        {
            get
            {
                return this._site;
            }
            set
            {
                this._site = value;
            }
        }

        #endregion IReportingEngineHost implementation

        /// <summary>
        /// Determines the ExportType from the file name.
        /// </summary>
        /// <param name="exportFile">The path of the file.</param>
        /// <returns>The Export Type.</returns>
        private ExportType ExportTypeFromFileName(string exportFile)
        {
            string extension = System.IO.Path.GetExtension(exportFile).ToLower();

            if (extension.Length > 1)
            {
                extension = extension.Substring(1);
                foreach (string val in Enum.GetNames(typeof(ExportType)))
                {
                    if (val.ToLower() == extension)
                    {
                        return (ExportType)Enum.Parse(typeof(ExportType), val, true); 
                    }
                }
            }
            throw new Exception(string.Format("Cannot determine the export file format type from [{0}].", exportFile));
        }

        /// <summary>
        /// Prints a report to a specific printer or exports a report to a file.
        /// </summary>
        /// <param name="exportFilePath">The file to export the report to, can be empty or null if the report is being printed.</param>
        /// <param name="exportService">The IExportService i.e. this object.</param>
        /// <param name="flags">Should always be set to SuppressUserInteraction.</param>
        /// <param name="exportType">The type of file if the report is being exported or prn if being printed.</param>
        /// <param name="printerName">The local printer name if the report is being printed.</param>
        /// <returns>True...not sure of the signifigance as it's undocumented.</returns>
        private bool PrintReport(string exportFilePath, IExportService exportService, ExportFlags flags, ExportType exportType, string printerName)
        {
            //Set the physical printer.
            if (exportType == ExportType.prn && !string.IsNullOrEmpty(printerName))
            {
                Type exporterType = exportService.GetType();
                PropertyInfo reportsPropertyInfo = exporterType.GetProperty("Reports", System.Reflection.BindingFlags.Public | System.Reflection.BindingFlags.Instance);
                IList reports = reportsPropertyInfo.GetValue(exportService, null) as IList;
                //Should never be zero, but just in case.
                if (reports.Count == 0)
                {
                    throw new Exception("No reports in the layout file");
                }
                //Get the ReportBatchItem
                var batchItem = reports[0];
                //Get the report.
                var reportPropertyInfo = batchItem.GetType().GetProperty("Report");
                //Get the actual report object.
                Sage.Reporting.Model.IReport report = reportPropertyInfo.GetValue(batchItem, null) as Sage.Reporting.Model.IReport;
                //Get the printer settings

                //Set the printer.
                //Sometimes the PrinterSettings can be null (usually when the report is saved with a printer which is not installed/available under the user profile 
                //which the report is being run.
                if (report.PaperLayout.PrinterSettings == null)
                {
                    System.Drawing.Printing.PrinterSettings printSettings = new System.Drawing.Printing.PrinterSettings();
                    //Set the printer name before setting the object to the PrinterSettings property. 
                    //Setting the object to the property fails, with the property remaining null.
                    printSettings.PrinterName = printerName;
                    report.PaperLayout.PrinterSettings = printSettings;
                }
                else
                {
                    System.Drawing.Printing.PrinterSettings printerSettings = report.PaperLayout.PrinterSettings;
                    printerSettings.PrinterName = printerName;
                }

            }
            else
            {
#if !SAGE2010
                flags |= ExportFlags.UseDefaultPrinter;
#endif
            }

            return exportService.Run(exportType, exportFilePath, flags);
        }

        /// <summary>
        /// Provides a generic means to build the dataset required for transient style reports.
        /// </summary>
        /// <param name="reportData">The data.</param>
        /// <param name="catalogName">The report name.</param>
        /// <returns>The converted dataset.</returns>
        private IList BuildTransientData(Sage.ObjectStore.MetaDataObjectCollection reportData, string catalogName)
        {
            IList result = null;
            Type catalogType = GetCatalogType(catalogName);
            Type dataListType = typeof(Sage.Accounting.Reporting.TransientDataList<>);

            result = Activator.CreateInstance(dataListType.MakeGenericType(new Type[] { catalogType })) as IList;

            foreach (object d in reportData)
            {
                object dataObj = Activator.CreateInstance(catalogType, new object[] { d });
                result.Add(dataObj);
            }
            return result;
        }

        /// <summary>
        /// Gets the data type of a transient report by catalog name.
        /// </summary>
        /// <param name="catalogName">The report name.</param>
        /// <returns>The type of data used in the transient report.</returns>
        private Type GetCatalogType(string catalogName)
        {
            System.Reflection.MethodInfo mi = typeof(Sage.Accounting.Reporting.ReportHelper).GetMethod("GetCatalogType", System.Reflection.BindingFlags.NonPublic | System.Reflection.BindingFlags.Static);
            Type result = mi.Invoke(null, new object[] { catalogName }) as Type;
            return result;
        }
    }
}
