﻿using System;
using System.Collections.Generic;
using System.Threading;

namespace Sage200ThreadLogon
{
    /// <summary>
    /// Marshals print requests.
    /// </summary>
    class ReportQueueManager
    {
        //Declare two queues for inter thread communication.
        //Set their maxiumum sizes to 1 so that requests are processed one at a time.
        private static BlockingQueue<ReportRequest> _requestQueue = new BlockingQueue<ReportRequest>(1);
        private static BlockingQueue<ReportReply> _replyQueue = new BlockingQueue<ReportReply>(1);

        //Static object used when starting up and shutting down the printer thread.
        private static object _syncLock = new object();

        //The thread used to print the report.
        private static Thread _reportProcessor;
        private static int _referenceCount;

        public ReportQueueManager()
        {
        }

        /// <summary>
        /// Starts and create the print thread.
        /// </summary>
        public static void Start()
        {
            lock (_syncLock)
            {
                System.Diagnostics.Debug.WriteLine("Report Processor Try Start - Ref Cnt - " + _referenceCount);
                if (_referenceCount == 0)
                {
                    System.Diagnostics.Debug.WriteLine("Report Processor ACTUAL Start - Ref Cnt - " + _referenceCount);

                    _reportProcessor = new Thread(ProcessReport);
                    //Important to set the state to STA as the Sage reporting engine
                    //interacts with the windows message queue.
                    _reportProcessor.SetApartmentState(ApartmentState.STA);
                    _reportProcessor.Start();
                }
                ++_referenceCount;
            }
        }

        /// <summary>
        /// Shuts down the manager.
        /// </summary>
        public static void Shutdown()
        {
            lock (_syncLock)
            {
                System.Diagnostics.Debug.WriteLine("Report Processor Try Shutdown - Ref Cnt - " + _referenceCount);
                --_referenceCount;
                if (_referenceCount == 0)
                {
                    System.Diagnostics.Debug.WriteLine("Report Processor ACTUAL Shutdown - Ref Cnt - " + _referenceCount);
                    _reportProcessor.Abort();
                    _reportProcessor = null;
                }
            }
        }

        /// <summary>
        /// Runs a print request.
        /// </summary>
        /// <param name="request">The print request to be run.</param>
        public static void RunReport(ReportRequest request)
        {
            //Enqueue the request. This call will block
            //if there is already an item in the queue.
            _requestQueue.Enqueue(request);

            //This call will block until a reply is received.
            ReportReply reply = _replyQueue.Dequeue();
            //If the request threw an exception, it is necessary to re-throw back to the client.
            if (reply.Success == ReplySuccess.Exception)
            {
                throw reply.Exception;
            }
        }

        /// <summary>
        /// Worker method to process ReportRequests
        /// </summary>
        private static void ProcessReport()
        {
            //Run forever!
            while (true)
            {
                //This method will block until a request is placed onto the queue.
                ReportRequest request = _requestQueue.Dequeue();
                //Set this thread's principal to the principal of the request, 
                //which the reporting engine then uses.
                Thread.CurrentPrincipal = request.Principal;

                System.Diagnostics.Debug.WriteLine(request.ThreadId + " - Report Processor Dequeue - " + request.ReportFileName);

                //Run the report. If an exception occurs whilst printer catch it and set the reply object accordingly
                //otherwise indicate success and enqueue the reply object.
                try
                {
                    ReportExporter report = new ReportExporter(request.ReportFileName, request.Criteria, request.ExportFilePath, request.Print, request.PrinterName);
                }
                catch (Exception ex)
                {
                    _replyQueue.Enqueue(new ReportReply() { Success = ReplySuccess.Exception, Exception = ex });
                }
                _replyQueue.Enqueue(new ReportReply() { Success = ReplySuccess.Success });
            }
        }
    }
}
